@@ -1,5 +1,6 @@
 import csv
 import os.path
+from SpecularLang.SpecLangTypes import Term, Operation, Type, SpecHelper, Operator, UnaryOperation
 
 from SpecularLang.SpecLangParser import SpecLangParser
 from SpecularLang.SpecLangVisitor import SpecLangVisitor
@@ -16,32 +17,16 @@ class SpecLangWalker(SpecLangVisitor):
         self.scenes = scenes
         self.talkative = talkative
 
-    operators = {
-        '==': ['ID', 'Number', 'String', 'Bool', 'None'],
-        '!=': ['ID', 'Number', 'String', 'Bool', 'None'],
-        '+': ['ID', 'Number', 'String'],
-        '-': ['ID', 'Number'],
-        '*': ['ID', 'Number'],
-        '/': ['ID', 'Number'],
-        '%': ['ID', 'Number'],
-        '>': ['ID', 'Number'],
-        '<': ['ID', 'Number'],
-        '>=': ['ID', 'Number'],
-        '<=': ['ID', 'Number'],
-        'and': ['ID', 'Bool'],
-        'or': ['ID', 'Bool'],
-        'not': ['ID', 'Bool'],
-    }
+
 
     def visitChoice(self, ctx:SpecLangParser.ChoiceContext):
-        #'[' STRING (',' STRING)*? ']' #choice stores a Number in $
         choices = {}
         i = 0
         while ctx.STRING(i) is not None:
             choices["choice{}".format(i)] = str(ctx.STRING(i)).strip('"')
             i += 1
         self.add_row([self.rowNum, "Choice", choices])
-        return {'type': 'ID', 'value': '${}'.format(self.rowNum - 1)}
+        return Term(Type.ID, '${}'.format(self.rowNum - 1))
 
     def visitDialog(self, ctx: SpecLangParser.DialogContext):
         if ctx.emotion():
@@ -55,103 +40,70 @@ class SpecLangWalker(SpecLangVisitor):
     def visitEmotion(self, ctx: SpecLangParser.EmotionContext):
         return ctx.getText().strip("(").strip(")").strip('"')
 
-    def visitTerm(self, ctx: SpecLangParser.TermContext):
+    def visitTerm(self, ctx: SpecLangParser.TermContext) -> Term:
         if ctx.NUMBER():
-            type_str = 'Number'
+            _type = Type.NUMBER
             value = ctx.getText().strip('"')
         elif ctx.TRUE() or ctx.FALSE():
-            type_str = 'Bool'
+            _type = Type.BOOL
             value = ctx.getText().strip('"')
         elif ctx.STRING():
-            type_str = 'String'
+            _type = Type.STRING
             value = self.convert_to_specular_string_format(ctx.getText())
         elif ctx.ID():
-            type_str = 'ID'
+            _type = Type.ID
             value = ctx.getText().strip('"')
         else:
-            type_str = 'None'
+            _type = Type.NONE
             value = ctx.getText().strip('"')
-        return {'type': type_str, 'value': value}
+        return Term(_type, value)
 
-    def visitAssignment(self, ctx:SpecLangParser.AssignmentContext):
+    def visitAssignment(self, ctx: SpecLangParser.AssignmentContext):
         if ctx.GLOBAL():
             is_global = 'Yes'
         else:
             is_global = 'No'
-        expr = self.visit(ctx.expression())
-        self.add_row([self.rowNum, "Assign", {'global': is_global, 'ID': str(ctx.ID()), 'type': expr['type'], 'assignment': expr['value']}], is_prescene=self.is_prescene)
+        term = SpecHelper.to_term(self.visit(ctx.expression()))
+        self.add_row([self.rowNum, "Assign", {'global': is_global, 'ID': str(ctx.ID()), 'type': term.type.value, 'assignment': term.value}], is_prescene=self.is_prescene)
 
-    def visitMult(self, ctx:SpecLangParser.MultContext):
-        term_0 = self.visit(ctx.expression(0))
-        term_1 = self.visit(ctx.expression(1))
-        expr_op = str(ctx.getChild(1))
-        # Assert that the operation is valid
-        assert term_0['type'] in self.operators[expr_op] and term_1['type'] in self.operators[expr_op], "{} {} {} is not a valid operation".format(term_0['type'], expr_op, term_1['type'])
-        if term_0['type'] == 'ID' or term_1['type'] == 'ID':
-            self.add_row([self.rowNum, "Expression", {'operator': expr_op, 'x': term_0['value'], 'y': term_1['value']}])
-            return {'type': 'ID', 'value': '${}'.format(self.rowNum - 1)}
-        else:  # Return * or / depending on the operator
-            return {'type': 'Number', 'value': int(term_0['value']) * int(term_1['value']) if expr_op == '*' else int(term_0['value']) / int(term_1['value'])}
+    def visitAdd(self, ctx: SpecLangParser.AddContext):
+        return self.perform_expression(ctx.expression(0), ctx.expression(1), Operator(str(ctx.getChild(1))))
 
-    def visitAdd(self, ctx:SpecLangParser.AddContext):
-        term_0 = self.visit(ctx.expression(0))
-        term_1 = self.visit(ctx.expression(1))
-        expr_op = str(ctx.getChild(1))
-        # Assert that the operation is valid
-        assert term_0['type'] in self.operators[expr_op] and term_1['type'] in self.operators[expr_op], "{} {} {} is not a valid operation".format(term_0['type'], expr_op, term_1['type'])        # type is string if term 0 or term 1 are strings else the type is Number
-        typeStr = "String" if term_0['type'] == 'String' or term_1['type'] == 'String' else "Number"
-        if term_0['type'] == 'ID' or term_1['type'] == 'ID':
-            self.add_row([self.rowNum, "Expression", {'operator': expr_op, 'x': term_0['value'], 'y': term_1['value']}])
-            return {'type': 'ID', 'value': '${}'.format(self.rowNum - 1)}
-        else:  # Return + or - depending on the operator
-            if typeStr == "String":
-                return {'type': 'String', 'value': self.append_to_formatted_string(term_0['value'],term_1['value'])}
-            else:
-                return {'type': 'Number', 'value': str(int(term_0['value']) + int(term_1['value'])) if expr_op == '+' else str(int(term_0['value']) - int(term_1['value']))}
+    def visitMult(self, ctx: SpecLangParser.MultContext):
+        return self.perform_expression(ctx.expression(0), ctx.expression(1), Operator(str(ctx.getChild(1))))
 
-    def visitEqual(self, ctx:SpecLangParser.EqualContext):
-        term_0 = self.visit(ctx.expression(0))
-        term_1 = self.visit(ctx.expression(1))
-        expr_op = str(ctx.getChild(1))
-        # Assert that the operation is valid
-        assert term_0['type'] in self.operators[expr_op] and term_1['type'] in self.operators[expr_op], "{} {} {} is not a valid operation".format(term_0['type'], expr_op, term_1['type'])
-        if term_0['type'] == 'ID' or term_1['type'] == 'ID':
-            self.add_row([self.rowNum, "Expression", {'operator': expr_op, 'x': term_0['value'], 'y': term_1['value']}])
-            return {'type': 'ID', 'value': '${}'.format(self.rowNum - 1)}
+    def perform_expression(self, expression_0, expression_1, operator: Operator):
+        term_0 = SpecHelper.to_term(self.visit(expression_0))
+        term_1 = SpecHelper.to_term(self.visit(expression_1))
+        operation = Operation(term_0, term_1, operator)
+        if operation.is_either_operand_of_type_id():
+            self.add_row([self.rowNum, "Expression", {'operator': operator.value, 'x': term_0.value, 'y': term_1.value}])
+            return Term(Type.ID, '${}'.format(self.rowNum - 1))
         else:
-            return {'type': 'Bool', 'value': term_0['value'] == term_1['value'] if expr_op == '==' else term_0['value'] != term_1['value']}
+            return operation.perform()
 
-    def visitAnd(self, ctx:SpecLangParser.AndContext):
-        term_0 = self.visit(ctx.expression(0))
-        term_1 = self.visit(ctx.expression(1))
-        assert term_0['type'] in self.operators['and'] and term_1['type'] in self.operators['and'], "{} {} {} is not a valid operation".format(term_0['type'], 'and', term_1['type'])
-        if term_0['type'] == 'ID' or term_1['type'] == 'ID':
-            self.add_row([self.rowNum, "Expression", {'operator': 'and', 'x': term_0['value'], 'y': term_1['value']}])
-            return {'type': 'ID', 'value': '${}'.format(self.rowNum - 1)}
+    def perform_unary_operation(self, expression, operator: Operator):
+        term = SpecHelper.to_term(self.visit(expression))
+        operation = UnaryOperation(self.visit(expression), operator)
+        if term.type == Type.ID:
+            self.add_row([self.rowNum, "Unary", {'operator': operator.value, 'x': term.value}])
+            return Term(Type.ID, '${}'.format(self.rowNum - 1))
         else:
-            return {'type': 'Bool', 'value': term_0['value'] and term_1['value']}
+            return operation.perform()
 
-    def visitOr(self, ctx:SpecLangParser.OrContext):
-        term_0 = self.visit(ctx.expression(0))
-        term_1 = self.visit(ctx.expression(1))
-        assert term_0['type'] in self.operators['or'] and term_1['type'] in self.operators['and'], "{} {} {} is not a valid operation".format(term_0['type'], 'or', term_1['type'])
-        if term_0['type'] == 'ID' or term_1['type'] == 'ID':
-            self.add_row([self.rowNum, "Expression", {'operator': 'or', 'x': term_0['value'], 'y': term_1['value']}])
-            return {'type': 'ID', 'value': '${}'.format(self.rowNum - 1)}
-        else:
-            return {'type': 'Bool', 'value': term_0['value'] or term_1['value']}
+    def visitEqual(self, ctx: SpecLangParser.EqualContext):
+        return self.perform_expression(ctx.expression(0), ctx.expression(1), Operator(str(ctx.getChild(1))))
 
-    def visitUnary(self, ctx:SpecLangParser.UnaryContext):
-        term = self.visit(ctx.expression())
-        expr_op = str(ctx.getChild(0))
-        assert term['type'] in self.operators[expr_op], "{} cannot be used with type: {}".format(expr_op, term['type'])
-        if term['type'] == 'ID':
-            self.add_row([self.rowNum, "Unary", {'operator': expr_op, 'x': term['value']}])
-            return {'type': 'ID', 'value': '${}'.format(self.rowNum - 1)}
-        else:
-            return {'type': "Bool" if expr_op == 'not' else "Number", 'value': str(not self.to_bool(term['value'])) if expr_op == 'not' else str(-int(term['value']))}
+    def visitAnd(self, ctx: SpecLangParser.AndContext):
+        return self.perform_expression(ctx.expression(0), ctx.expression(1), Operator(str(ctx.getChild(1))))
 
-    def visitParen(self, ctx:SpecLangParser.ParenContext):
+    def visitOr(self, ctx: SpecLangParser.OrContext):
+        return self.perform_expression(ctx.expression(0), ctx.expression(1), Operator(str(ctx.getChild(1))))
+
+    def visitUnary(self, ctx: SpecLangParser.UnaryContext):
+        return self.perform_unary_operation(ctx.expression(), Operator(str(ctx.getChild(0))))
+
+    def visitParen(self, ctx: SpecLangParser.ParenContext):
         return self.visit(ctx.expression())
 
     def visitScene_statement(self, ctx:SpecLangParser.Scene_statementContext):
@@ -166,20 +118,20 @@ class SpecLangWalker(SpecLangVisitor):
             self.rowNum = 0
             self.is_prescene = True
 
-    def visitIfstatement(self, ctx:SpecLangParser.IfstatementContext):
+    def visitIfstatement(self, ctx: SpecLangParser.IfstatementContext):
         current_row = self.rowNum
-        term = self.visit(ctx.expression())
+        term = SpecHelper.to_term(self.visit(ctx.expression()))
 #        elseifs = ctx.else_if_statement()
         elseifs = []
         else_state = ctx.else_statement()
         elses_exist = elseifs != [] or else_state is not None
-        if term['value'] == 'True':
+        if str(term.value) == 'True':
             self.visit(ctx.block())
-        elif term['value'] == 'False' and not elses_exist:
+        elif str(term.value) == 'False' and not elses_exist:
             return
         else:
             elif_count = 0
-            self.add_row([self.rowNum, "If", {'condition': term['value'], 'jump': 'endIf_{}'.format(current_row) if not elses_exist else 'elseif_{}'.format(current_row)}])
+            self.add_row([self.rowNum, "If", {'condition': term.value, 'jump': 'endIf_{}'.format(current_row) if not elses_exist else 'elseif_{}'.format(current_row)}])
             self.visit(ctx.block())
             if elses_exist:
                 self.add_row([self.rowNum, "JumpToLabel", {'name': 'endIf_{}'.format(current_row)}])
@@ -206,14 +158,14 @@ class SpecLangWalker(SpecLangVisitor):
     def visitWhileLoop(self, ctx:SpecLangParser.WhileLoopContext):
         current_row = self.rowNum
         self.add_row([self.rowNum, "Label", {'name': 'beginWhile_{}'.format(current_row)}])
-        term = self.visit(ctx.expression())
-        if term['value'] == 'True':
+        term = SpecHelper.to_term(self.visit(ctx.expression()))
+        if str(term.value) == 'True':
             raise Exception("While loop around line: {} will run forever (which is not allowed)".format(current_row))
-        elif term['value'] == 'False':
+        elif str(term.value) == 'False':
             self.remove_last_row()
             return
         else:
-            self.add_row([self.rowNum, "While", {'condition': term['value'], 'jump': 'endWhile_{}'.format(current_row)}])
+            self.add_row([self.rowNum, "While", {'condition': term.value, 'jump': 'endWhile_{}'.format(current_row)}])
             self.visit(ctx.block())
             self.add_row([self.rowNum, "JumpToLabel", {'name': 'beginWhile_{}'.format(current_row)}])
             self.add_row([self.rowNum, "Label", {'name': 'endWhile_{}'.format(current_row)}])
@@ -232,14 +184,9 @@ class SpecLangWalker(SpecLangVisitor):
 #            self.visit(ctx.block())
 #            self.add_row([self.rowNum, "While", {'condition': term['value'], 'jump': 'doWhile_{}'.format(current_row)}])
 
-    def visitCustomDirection(self, ctx:SpecLangParser.CustomDirectionContext):
-        self.add_row([self.rowNum, "Custom", {'name': str(ctx.STRING()).strip('"')}])
+    def visitStage_direction(self, ctx:SpecLangParser.Stage_directionContext):
+        self.add_row([self.rowNum, "StageDir", {'name': str(ctx.STRING()).strip('"')}])
 
-    def visitFadeIn(self, ctx:SpecLangParser.FadeInContext):
-        self.add_row([self.rowNum, "FadeIn", {}])
-
-    def visitFadeOut(self, ctx:SpecLangParser.FadeOutContext):
-        self.add_row([self.rowNum, "FadeOut", {}])
 
 # region: Utils
     def to_bool(self, string: str):
